"""
Modules contains set of helpers to work with resource selector.

Copyright 2022 nMachine.io
"""
from __future__ import annotations

from typing import Any, List, Optional, Tuple


def _convert_kv_to_exp(
    kv_list: List[Tuple[str, str]], incl: Optional[bool] = True
) -> str:
    """Converts list of tuples to string in format
    k=v or k!=v separated with comma.

    TODO: use operator in or notin for the same "k"
    """
    rets = []
    sep = "=" if incl else "!="
    for name, value in kv_list:
        rets.append(f"{name}{sep}{value}")
    return ",".join(rets)


def join_not_empty(list_to_join: List[str]) -> str:
    """Joins list of string and ignores empty items."""
    return ",".join(item for item in list_to_join if item)


def convert_selectors(query: Any) -> None:
    """Converts fields/not_fields and labels/not_labels
    to field_selector and label_selector."""

    field_selector = [query.get("field_selector", "")]
    if query.get("fields"):
        field_selector.append(_convert_kv_to_exp(query["fields"]))
        del query["fields"]
    if query.get("not_fields"):
        field_selector.append(_convert_kv_to_exp(query["not_fields"], False))
        del query["not_fields"]
    if field_selector:
        field_selector_txt = join_not_empty(field_selector)
        if field_selector_txt:
            query["field_selector"] = field_selector_txt

    label_selector = [query.get("label_selector", "")]
    if query.get("labels"):
        label_selector.append(_convert_kv_to_exp(query["labels"]))
        del query["labels"]
    if query.get("not_labels"):
        label_selector.append(_convert_kv_to_exp(query["not_labels"], False))
        del query["not_labels"]
    if label_selector:
        label_selector_txt = join_not_empty(label_selector)
        if label_selector_txt:
            query["label_selector"] = label_selector_txt



"""
Copyright 2022 nMachine.io

Requirements:
- e2e cluster, configuration from KUBECONFIG or ~/.kube/config
- applied configuration from k8kat/e2e_tests/cluster-e2e.yaml

"""

import unittest

from k8kat import ServiceAccount


class TestServiceAccount(unittest.TestCase):
    """Set of e2e tests for ServiceAccount resource."""

    def test_list_sa_from_e2e_namespace(self) -> None:
        """Tests listing Service Accounts from 'e2e' namespace"""
        ssa_list = ServiceAccount.list(ns="e2e")
        self.assertEqual(len(ssa_list), 2)
        self.assertEqual({ssa.name for ssa in ssa_list}, {"default", "example-sa"})

    def test_list_sa_from_e2e_namespace_excluding_sys(self) -> None:
        """Tests listing Service Accounts from 'e2e' namespace without default sa."""
        ssa_list = ServiceAccount.list_excluding_sys(ns="e2e")
        self.assertEqual(len(ssa_list), 1)
        self.assertEqual(ssa_list[0].name, "example-sa")

    def test_find_sa_with_not_found_error(self) -> None:
        """Tests finding not-existing service account"""
        sec = ServiceAccount.find("sa-does-not-exist", ns="e2e")
        self.assertIsNone(sec)

    def test_find_sa_with_secrets(self) -> None:
        """Tests finding service account"""
        sec = ServiceAccount.find("example-sa", ns="e2e")
        assert sec

        self.assertEqual(sec.name, "example-sa")
        self.assertEqual(sec.labels, {})
        secrets = sec.secrets()
        self.assertEqual(len(secrets), 1)
        self.assertEqual(secrets[0].name, "example-credentials")



"""
Copyright 2022 nMachine.io

Requirements:
- e2e cluster, configuration from KUBECONFIG or ~/.kube/config
- applied configuration from k8kat/e2e_tests/cluster-e2e.yaml

"""

import unittest

from k8kat import Secret


class TestSecret(unittest.TestCase):
    """Set of e2e tests for Secret resource."""

    def test_list_secret_from_e2e_namespace(self) -> None:
        """Tests listing pods from 'e2e' namespace"""
        secs = Secret.list(ns="e2e")
        self.assertEqual(len(secs), 1)

    def test_find_secret_with_not_found_error(self) -> None:
        """Tests finding not-existing secret"""
        sec = Secret.find("secret-does-not-exist", ns="e2e")
        self.assertIsNone(sec)

    def test_find_secret_with_data(self) -> None:
        """Tests finding secret and get decoded data"""
        sec = Secret.find("example-credentials", ns="e2e")
        assert sec
        self.assertEqual(
            sec.decoded_data(), {"PASSWORD": "1f2d1e2e67df", "USER_NAME": "admin"}
        )



"""
Copyright 2022 nMachine.io

Provides the decorator @classproperty.
"""


class ClassPropertyDescriptor:
    """Implements getter/setter for class property"""

    def __init__(self, fget, fset=None):  # type: ignore
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):  # type: ignore
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):  # type: ignore
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):  # type: ignore
        """Setter for properties"""
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):  # type: ignore
    """Decorator for class properties"""
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)



"""
Copyright 2022 nMachine.io

Set of functions to work with units.
"""

from typing import Dict, Optional, Tuple, Union


def find_multiplier_mapping(expr: str) -> Optional[Tuple[str, Union[int, float]]]:
    """Finds the correct multiplier for unit conversion."""
    if expr is None:
        return None
    for candidate, multiplier in unit_map.items():
        if expr.endswith(candidate):
            return candidate, multiplier
    return None


def parse_quant_expr(src_expr: str, target: str = "") -> Optional[float]:
    """Parses the source expression into standard units
    or target units (if specified)."""
    if src_expr is not None:
        src_mapping = find_multiplier_mapping(src_expr)
        target_mapping = find_multiplier_mapping(target)
        if not src_mapping or not target_mapping:
            return None

        try:
            src_quantity, src_multiplier = src_mapping
            _, target_multiplier = target_mapping
            unitless = float(src_expr.strip(src_quantity))
            return unitless * src_multiplier / target_multiplier
        except ValueError:
            return None
    else:
        return None


def humanize_cpu_quant(cores: float, with_unit: bool = False) -> str:
    """Returns CPU quantity in cores rounded to single decimal."""
    return f"{cores:.1f} {'Cores' if with_unit else ''}".strip(" ")


def mem_quant_mult(byte_value: float) -> Tuple[Optional[float], Optional[str]]:
    """Returns memory quantity converted from bytes to higher level units.
    Automatically picks the right units for friendliest display."""
    unit_map_items = list(unit_map.items())
    sorted_items = sorted(unit_map_items, key=lambda item: item[1])
    unit, bytes_in_unit = None, None
    for (try_unit, multiplier) in sorted_items:
        bytes_in_unit = byte_value / multiplier
        if try_unit and bytes_in_unit < 1000:
            unit = try_unit
            break
    return bytes_in_unit, unit


def humanize_mem_quant(byte_value: float) -> Optional[str]:
    """Returns memory quantity converted from bytes to higher level units.
    Automatically picks the right units for friendliest display."""
    bytes_in_unit, unit = mem_quant_mult(byte_value)
    return f"{int(bytes_in_unit)}{unit}b" if bytes_in_unit else None


unit_map: Dict[str, Union[int, float]] = {
    "Ei": 2**60,
    "Pi": 2**50,
    "Ti": 2**40,
    "Gi": 2**30,
    "Mi": 2**20,
    "Ki": 2**10,
    "E": 10**18,
    "P": 10**15,
    "T": 10**12,
    "G": 10**9,
    "M": 10**6,
    "K": 10**3,
    "k": 10**3,  # k8s uses both K and k for "kilo"
    "m": 10**-3,
    "micro": 10**-6,
    "n": 10**-9,
    "p": 10**-12,
    "": 1,
}



"""
Copyright 2022 nMachine.io
"""


from __future__ import annotations

from typing import Any, List, Optional

from kubernetes.client import models  # type: ignore

from k8kat.broker import Broker
from k8kat.utils import classproperty

from . import BaseResource
from .secret import Secret


class ServiceAccount(BaseResource):
    """Service Account resource.

    This class inherits from BaseResource and provide:

    1. static methods to list/find/delete service accounts using K8s API
    2. object properties, methods to get stored data
    """

    def __init__(self, raw: models.V1ConfigMap, broker: Broker):
        BaseResource.__init__(self, raw, broker)

    def body(self) -> models.V1ServiceAccount:
        """Returns raw object V1ServiceAccount"""
        return self.raw

    def secrets(self) -> List[Secret]:
        """Returns list of the secrets in the same namespace that
        pods running using this ServiceAccount are allowed to use."""
        secrets_raw = self.body().secrets or []
        secrets = []
        for secret_raw in secrets_raw:
            secret = Secret.find(
                secret_raw.name, ns=self.raw.metadata.namespace, broker=self.broker
            )
            if secret:
                secrets.append(secret)
        return secrets

    @classmethod
    def list_excluding_sys(
        cls, ns: str, broker: Optional[Broker] = None, **query: Any
    ) -> List[ServiceAccount]:
        """Returns service accounts without defaults."""
        not_fields = list(query.get("not_fields", []))
        not_fields.append(("metadata.name", "default"))
        query["not_fields"] = not_fields
        return cls.list(ns, broker=broker, **query)

    @classmethod
    def k8s_list_method(
        cls, broker: Broker, **kwargs: Any
    ) -> models.V1ServiceAccountList:
        ns = kwargs["ns"]
        del kwargs["ns"]
        return broker.core_v1_api.list_namespaced_service_account(
            namespace=ns, **kwargs
        )

    @classmethod
    def k8s_read_method(cls, broker: Broker, **kwargs: Any) -> models.V1ServiceAccount:
        ns = kwargs.get("ns")
        del kwargs["ns"]
        return broker.core_v1_api.read_namespaced_service_account(
            namespace=ns, **kwargs
        )

    @classmethod
    def k8s_patch_method(cls, broker: Broker, **kwargs: Any) -> models.V1ServiceAccount:
        return broker.core_v1_api.patch_namespaced_service_account(**kwargs)

    @classmethod
    def k8s_delete_method(cls, broker: Broker, **kwargs: Any) -> models.V1Status:
        return broker.core_v1_api.delete_namespaced_service_account(**kwargs)

    @classproperty
    def kind(self) -> str:
        return "ServiceAccount"



"""
Helper methods for Pod.

Copyright 2022 nMachine.io
"""


from http.client import HTTPResponse
from io import BytesIO
from typing import IO, Any, Callable, List, Optional, Union
from urllib.parse import unquote_plus

from kubernetes.client import V1Container  # type: ignore

from k8kat.utils import units


def coerce_cmd_format(cmd: Union[str, List[str]]) -> List[str]:
    """Returns command as list (coverts from string if necessary)"""
    if isinstance(cmd, str):
        parts = cmd.split(" ")
        # parts = [part.replace("SPACE_CHAR", " ") for part in parts]
        parts = [unquote_plus(part) for part in parts]
        return parts
    return cmd


def container_req_or_lim(
    container: V1Container, metric: str, resource: str
) -> Optional[float]:
    """Gets container capacity and returns in cores (cpu) / bytes (memory)."""
    assert metric in ["requests", "limits"]
    assert resource in ["cpu", "memory", "ephemeral-storage"]
    metrics_dict = getattr(container.resources, metric, None)
    capacity_expr = (metrics_dict or {}).get(resource, None)
    return capacity_expr and units.parse_quant_expr(capacity_expr)


def build_curl_cmd(**params: Any) -> List[str]:
    """Builds curl command line"""
    raw_headers = params.get("headers", {})
    headers = [f"{0}: {1}".format(k, v) for k, v in raw_headers]
    body = params.get("body", None)

    cmd = [
        "curl" if params.get("with_command") else None,
        "-s",
        "-i",
        "-X",
        params.get("verb", "GET"),
        "-H" if headers else None,
        headers if headers else None,
        "-d" if body else None,
        body if body else None,
        "--connect-timeout",
        "1",
        f"{params['url']}{params.get('path', '/')}",
    ]
    return [part for part in cmd if part is not None]


def parse_response(response_str: str) -> Optional[HTTPResponse]:
    """Converts output from curl to HTTPResponse."""
    if response_str:

        class _FakeSocket:
            # pylint: disable=unused-argument
            # pylint: disable=missing-function-docstring
            def __init__(self, response_bytes: bytes):
                self._file = BytesIO(response_bytes)

            def makefile(self, *args: Any, **kwargs: Any) -> IO:
                return self._file

        source = _FakeSocket(response_str.encode("ascii"))
        # noinspection PyTypeChecker
        response = HTTPResponse(source)  # type: ignore
        response.begin()
        return response

    return None


def when_running_normally(func: Callable) -> Callable:
    """Note: needs to be defined at top of class."""

    def running_normally_func(*args: Any, **kwargs: Any) -> Any:
        if args[0].is_running_normally():
            return func(*args, **kwargs)
        return None

    return running_normally_func



"""
Copyright 2022 nMachine.io
"""


from __future__ import annotations

from typing import Any, Dict, List

from kubernetes.client import models  # type: ignore

from k8kat.broker import Broker
from k8kat.utils import classproperty

from . import BaseResource


class Ingress(BaseResource):
    """Ingress resource.

    This class inherits from BaseResource and provide:

    1. static methods to list/find/delete ingresses using K8s API
    2. object properties, methods to get ingress' rules
    """

    def __init__(self, raw: models.V1ConfigMap, broker: Broker):
        BaseResource.__init__(self, raw, broker)

    def body(self) -> models.V1Ingress:
        """Returns raw object V1Ingress"""
        return self.raw

    def basic_rules(self) -> Dict[str, List[Dict]]:
        """Returns dict with mapping hosts to their rules."""
        result = {}
        rules: List[models.V1IngressRule] = self.body().spec.rules
        for rule in rules:
            if rule.host:
                bundles = []
                ingress_paths: List[models.V1HTTPIngressPath] = rule.http.paths
                for ingress_path in ingress_paths:
                    bundles.append(
                        dict(
                            service=ingress_path.backend.service.name,
                            port=ingress_path.backend.service.port.number,
                            path=ingress_path.path,
                        )
                    )
                result[rule.host] = bundles
        return result

    def flat_rules(self) -> List[Dict]:
        """Returns rules as flat list."""
        as_list = []
        for host, rules in list(self.basic_rules().items()):
            for rule in rules:
                as_list.append({"host": host, **rule})
        return as_list

    @classmethod
    def k8s_list_method(cls, broker: Broker, **kwargs: Any) -> models.V1SecretList:
        ns = kwargs["ns"]
        del kwargs["ns"]
        return broker.networking_v1_api.list_namespaced_ingress(namespace=ns, **kwargs)

    @classmethod
    def k8s_read_method(cls, broker: Broker, **kwargs: Any) -> models.V1Secret:
        ns = kwargs.get("ns")
        del kwargs["ns"]
        return broker.networking_v1_api.read_namespaced_ingress(namespace=ns, **kwargs)

    @classmethod
    def k8s_patch_method(cls, broker: Broker, **kwargs: Any) -> models.V1Secret:
        return broker.networking_v1_api.patch_namespaced_ingress(**kwargs)

    @classmethod
    def k8s_delete_method(cls, broker: Broker, **kwargs: Any) -> models.V1Status:
        return broker.networking_v1_api.delete_namespaced_ingress(**kwargs)

    @classproperty
    def kind(self) -> str:
        return "Ingress"

    @classproperty
    def res_name_plural(self) -> str:
        return "ingresses"



"""
Copyright 2022 nMachine.io
"""


from __future__ import annotations

import json
from typing import Any, Dict, Optional

import yaml  # type: ignore
from kubernetes.client import models  # type: ignore

from k8kat.broker import Broker
from k8kat.utils import classproperty

from . import BaseResource


class ConfigMap(BaseResource):
    """ConfigMap resource.

    This class inherits from BaseResource and provide:

    1. static methods to list/find/patch/delete configmaps using K8s API
    2. object properties, methods to get stored data
    """

    def __init__(self, raw: models.V1ConfigMap, broker: Broker):
        BaseResource.__init__(self, raw, broker)

    @property
    def data(self) -> Dict[str, str]:
        """Returns key-value pairs stored in object"""
        return self.raw.data or {}

    @property
    def binary_data(self) -> Dict[str, str]:
        """Returns key-binary-value pairs stored in object"""
        return self.raw.binary_data or {}

    def jget(
        self, key: Optional[str] = None, default: Optional[Any] = None
    ) -> Optional[Any]:
        """Gets a value from ConfigMap. Values stored in ConfigMap has to
        be json encoded.

        Args:
            key: name of key in ConfigMap, default master
            default: default value

        Returns:
            value stored on key"""

        key = key or "master"
        raw_value = self.data.get(key)
        obtained = raw_value and json.loads(raw_value)
        return obtained if obtained is not None else default

    def yget(
        self, key: Optional[str] = None, default: Optional[Any] = None
    ) -> Optional[Any]:
        """Gets a value from ConfigMap. Values stored in ConfigMap has to
        be yaml encoded.

        Args:
            key: name of key in ConfigMap, default master
            default: default value

        Returns:
            value stored on key"""

        key = key or "master"
        raw_value = self.data.get(key)
        obtained = raw_value and yaml.load(raw_value, Loader=yaml.FullLoader)
        return obtained if obtained is not None else default

    def jpatch(
        self, content: Any, key: Optional[str] = None, merge: Optional[bool] = False
    ) -> None:
        """Modifies a value from ConfigMap.

        Args:
            content: new value to store
            key: name of key in ConfigMap, default master
            merge: if true it tries to merge values with values from the configmap.
                   if values are not dict it raises ValueError"""
        key = key or "master"
        if merge:
            current_content = self.jget(key)
            if isinstance(current_content, dict) and isinstance(content, dict):
                content = {**current_content, **content}
            else:
                raise ValueError("Only dicts can be merged")
        self.raw.data = {key: json.dumps(content)}
        return self.patch()

    @classmethod
    def k8s_list_method(cls, broker: Broker, **kwargs: Any) -> models.V1PodList:
        ns = kwargs["ns"]
        del kwargs["ns"]
        return broker.core_v1_api.list_namespaced_config_map(namespace=ns, **kwargs)

    @classmethod
    def k8s_read_method(cls, broker: Broker, **kwargs: Any) -> models.V1ConfigMap:
        ns = kwargs.get("ns")
        del kwargs["ns"]
        return broker.core_v1_api.read_namespaced_config_map(namespace=ns, **kwargs)

    @classmethod
    def k8s_patch_method(cls, broker: Broker, **kwargs: Any) -> models.V1ConfigMap:
        return broker.core_v1_api.patch_namespaced_config_map(**kwargs)

    @classmethod
    def k8s_delete_method(cls, broker: Broker, **kwargs: Any) -> models.V1Status:
        return broker.core_v1_api.delete_namespaced_config_map(**kwargs)

    @classproperty
    def kind(self) -> str:
        return "ConfigMap"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="k8kat",
    version="0.1.1",
    author="NMachine",
    author_email="xavier@nmachine.io",
    description="Sugar for Kubernetes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nmachine-cs/k8kat",
    packages=setuptools.find_packages(
        exclude=[
            "k8kat.tests.*",
            "k8kat.tests",
            "k8kat.e2e_test.*",
            "k8kat.e2e_test",
        ]
    ),
    install_requires=[
        "python-dotenv",
        "kubernetes",
        "typing-extensions",
        "inflection",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
